//
// Moving bitmaps screensaver
// ----------------------------
// (c) 1996 Lucian Wischik
// portion (c) 1991-1995 Borland International
//
//
// TBitmapsData is the name of the class that holds the screensaver's data.
// The bitmap is loaded when TBitmapsData is created in response to WM_CREATE,
// and is deleted when TBitmapsData is deleted in response to WM_DESTORY.
// Because LoadImage is used, we don't have to do any palette stuff ourselves.
// To check if we are running full-screen or preview, we look at GetParent(hwnd):
// if this is NULL, then we are full-screen; if not, then not.
// To play the sound, we have to check if the general tab is set to 'mute'
// or not. This is done with GetScrGeneralSettings(). If we had a preview
// box or we allowed ActiveConfig, we also would need a response under
// SCRM_CHANGE.


#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdlib.h>
#include <mmsystem.h>
#pragma hdrstop

#include "../../lib/scrplus.h"
#include "boing.rh"

HINSTANCE hInstance;


class TBitmapData { public:
  TBitmapData();
  ~TBitmapData();
  void Seed(HWND);
  void Step(HDC);
  //
  int RandomMax;
  int AbsoluteDelta;
  int XPosition;
  int YPosition;
  int XDelta;
  int YDelta;
  int XLimit;
  int YLimit;
  int XBit;
  int YBit;
  int XScreen;
  int YScreen;
  BOOL isfull;
  double xscale;
  double yscale;
  BOOL bounce;
  HDC hmemdc;
  HGDIOBJ holdbitmap;
  HBITMAP hbitmap;
  BOOL mute;
};

TBitmapData::TBitmapData()
{ hmemdc=0; holdbitmap=0; hbitmap=0;
  TScrGeneralSettings *scg=new TScrGeneralSettings;
  ScrGetGeneralSettings(scg);
  mute=scg->mute;
  delete scg;
}
TBitmapData::~TBitmapData()
{ if (hmemdc!=NULL)
  { if (hbitmap!=NULL) SelectObject(hmemdc,holdbitmap);
	 DeleteObject(hbitmap);
	 hbitmap=0; holdbitmap=0;
	 DeleteDC(hmemdc); hmemdc=0;
  }
}

void TBitmapData::Seed(HWND hwnd)
{
  RandomMax=4; AbsoluteDelta=7;
  XPosition = 0;
  YPosition = 0;
  XDelta = random(RandomMax) + AbsoluteDelta;
  YDelta = random(RandomMax) + AbsoluteDelta;
  XScreen=GetSystemMetrics(SM_CXSCREEN);
  YScreen=GetSystemMetrics(SM_CYSCREEN);
  RECT rc; GetWindowRect(hwnd,&rc);
  XLimit=rc.right-rc.left; YLimit=rc.bottom-rc.top;
  BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE;
  if (isfull)
  { xscale=1.0; yscale=1.0; }
  else
  { xscale=((double)XLimit)/((double)XScreen);
    yscale=((double)YLimit)/((double)YScreen);
  }
  HDC hdc=GetDC(hwnd);
  hmemdc=CreateCompatibleDC(hdc);
  ReleaseDC(hwnd,hdc);
#if (WINVER>=0x0400)
  hbitmap=(HBITMAP)LoadImage(hInstance,MAKEINTRESOURCE(BMP_BOING),IMAGE_BITMAP,0,0,0);
  BITMAP bmp;
  GetObject(hbitmap,sizeof(bmp),&bmp);
  XBit=bmp.bmWidth; YBit=bmp.bmHeight;
  holdbitmap=SelectObject(hmemdc,hbitmap);
#else
  hbitmap=NULL;
  XBit=0; YBit=0;
  holdbitmap=NULL;
#pragma message ("ERROR! This example DOESN'T WORK unless you have the new win32 headers.  ")
#pragma message ("You should get a newer version of the Windows SDK.  ")
#endif
}

void TBitmapData::Step(HDC hdc)
{   bounce=FALSE;
	 if (isfull)
		BitBlt(hdc,XPosition,YPosition,XBit,YBit,hmemdc,0,0,SRCCOPY);
	 else
		StretchBlt(hdc,XPosition*xscale,YPosition*yscale,XBit*xscale,YBit*yscale,hmemdc,0,0,XBit,YBit,SRCCOPY);
	 RECT rc;
	 if (YDelta<0)
	 { rc.top=YPosition+YDelta; rc.bottom=YPosition+YBit-YDelta; }
    else
    { rc.top=YPosition-YDelta; rc.bottom=YPosition+YBit+YDelta; }
    if (XDelta<0)
    { rc.left=XPosition+XBit; rc.right=rc.left-XDelta; }
    else
    { rc.left=XPosition-XDelta; rc.right=rc.left+XDelta; }
    rc.left*=xscale; rc.right*=xscale; rc.top*=yscale; rc.bottom*=yscale;
    FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
    if (XDelta<0)
    { rc.left=XPosition+XDelta; rc.right=XPosition+XBit-XDelta;
    }
    else
    { rc.left=XPosition-XDelta; rc.right=XPosition+XBit+XDelta;
    }
    if (YDelta<0)
    { rc.top=YPosition+YBit; rc.bottom=rc.top-YDelta; }
	 else
    { rc.top=YPosition-YDelta; rc.bottom=rc.top+YDelta; }
    rc.left*=xscale; rc.right*=xscale; rc.top*=yscale; rc.bottom*=yscale;
    FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));

    XPosition += XDelta;
    YPosition += YDelta;

    if (XPosition > XScreen-XBit) {
      XPosition = XScreen-XBit;
      XDelta = -(random(RandomMax) + AbsoluteDelta);
      bounce = TRUE;
    }
    if (XPosition < 0) {
      XPosition = 0;
      XDelta = random(RandomMax) + AbsoluteDelta;
      bounce = TRUE;
    }

    if (YPosition > YScreen-YBit) {
      YPosition = YScreen-YBit;
		YDelta = -(random(RandomMax) + AbsoluteDelta);
      bounce = TRUE;
    }

    if (YPosition < 0) {
      YPosition = 0;
      YDelta = random(RandomMax) + AbsoluteDelta;
      bounce = TRUE;
    }

    if (bounce && !mute) PlaySound(MAKEINTRESOURCE(WAV_BOING),hInstance,SND_RESOURCE | SND_ASYNC | SND_NODEFAULT);

}



LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TBitmapData *bd=(TBitmapData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (bd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg) {
    case (WM_CREATE):
		{ TBitmapData *bd=new TBitmapData();
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)bd);
        bd->Seed(hwnd);
      } break;
    case (WM_ERASEBKGND):
      { HDC hdc=(HDC) wParam; RECT rc; GetClientRect(hwnd,&rc);
        FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
      } break;
    case (SCRM_CHANGE):
      { TScrGeneralSettings *sgs=(TScrGeneralSettings *)lParam;
        bd->mute=sgs->mute;
      } break;
    case (SCRM_TIMER):
      { HDC hdc=GetDC(hwnd);
        bd->Step(hdc);
        ReleaseDC(hwnd,hdc);
        return 18;
		}
    case (WM_DESTROY):
      { SetWindowLong(hwnd,GWL_USERDATA,0);
        delete bd;
      } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  TScrPrefs *p=ScrCreatePrefs();
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);

  return ScrExecute(hInstance,MyScreenSaverProc);
}



