//
// Ghosts screensaver
// --------------------
// (c) 1996 Lucian Wischik
//
// I think that most of this is pretty much self explanatory...

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include "../../lib/ddraw.h"
#pragma hdrstop
#include <mmsystem.h>
#include <commdlg.h>
#include <stdio.h>
#include <process.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>
#include "../../lib/scrplus.h"
#include "bits.h"

#include "ghosts.rh"


HINSTANCE hInstance;



class TFileData
{ public:
  char name[MAX_PATH];
  char path[MAX_PATH];
  TFileData *next;
  TFileData(char *name,char *path,TFileData *);
  ~TFileData();
  void Insert(char *name,char *path);
};

TFileData::TFileData(char *a_name,char *a_path,TFileData *a_next) {strcpy(name,a_name); strcpy(path,a_path); next=a_next;}
TFileData::~TFileData() { if (next!=NULL) delete next; next=NULL; }
void TFileData::Insert(char *a_name, char *a_path)
{ next=new TFileData(name,path,next); strcpy(name,a_name); strcpy(path,a_path); }

class TGhostsData
{ public:
  TBits previewbits;   // for the dialog control, to show what it will look like
  bool ignorechanges;  // Do we ignore dialog changes at the moment?
  TFileData *files;    // linked list of bitmap filenames
  bool settingschanged;// Do the contents of the config dialog have to be rewritten?

  int Screen;          // Screen size selection in registry
  bool Slow;           // do we slow it down?

  int mode;            // 0=prep. 1=thread. 2=static. 3=fadein. 4=fadeout.
  TFileData *currentimage; // Which one we're on at the moment
  TBits facebits;      // the currently-being-used ghost image
  CRITICAL_SECTION facecritsec; // this protects the above two fields
  HANDLE donefaceevent; // Signalled when the face has been rendered
  HANDLE abortworkevent;// Signalled if we have to stop
  TBits bits;          // for the actual animation
  TBits backdropbits;  // to show while we're suspended
  bool issuspended;    // whether a dialog is up
  bool ischangingmode; // SetDisplayMode() causes spurious WM_ACTIVATE, which we must ignore
  bool hasjustchangedmode; // and it also posts a WM_MOUSEMOVE
  bool isfull;         // Are we in full-screen mode?
  bool wantddx;        // and therefore, do we want to use ddx?
  bool isstarted;      // Has the sim yet started?
  int frame;           // current frame number
  int cx,cy;           // screen sizes to use
  int period;          // timer delay between cycles
  unsigned long hthread;
  unsigned int startfadeintime;
  unsigned int startfadeouttime;
  unsigned int startnewimagetime;
  int screentop,imagetop,imagebottom,screenbottom;
  int screenleft,imageleft,imageright,screenright;
  int imagerestofline;


  TGhostsData(BOOL);
  ~TGhostsData();
  void WinStart(HWND);
  unsigned int WinTimer(HWND);
  void WinPrepare();      // to start a face-ghosting thread running
  void WinDoFrame();
  void WinStop(HWND);
  void WinSuspend(HWND);
  void WinResume(HWND);

  void ReadRegistry();
  void WriteRegistry();
  void SetControls(HWND);
  BOOL GetControls(HWND);
  void Add(HWND);
  void Remove(HWND);
  void SelChange(HWND);
  void ShowFiles(HWND,bool managechange=true);
};


TGhostsData::TGhostsData(BOOL a_isfull)
{ isfull=a_isfull;
  wantddx=false;
  isstarted=false;
  ischangingmode=false; issuspended=false;
  hasjustchangedmode=false;
  Screen=0;
  Slow=false;
  frame=0;
  period=1;
  hthread=0;
  cx=0; cy=0;
  files=NULL;
  settingschanged=false;
  InitializeCriticalSection(&facecritsec);
  donefaceevent=CreateEvent(NULL,TRUE,FALSE,NULL);
  abortworkevent=CreateEvent(NULL,TRUE,FALSE,NULL);
}
TGhostsData::~TGhostsData()
{ delete files;
  previewbits.Reset();
}


void TGhostsData::ReadRegistry()
{ if (files!=NULL) delete files;
  char inpath[100];
  wsprintf(inpath,"<internal>%i",BMP_LU);
  files=new TFileData("Lucian Wischik",inpath,NULL);
  LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Ghosts",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) {settingschanged=true; return; }
  // 1. Screen
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Screen",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Screen=(BYTE)val;
  if (Screen<0) Screen=0; if (Screen>5) Screen=5;
  // 3. Slow
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Slow",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Slow=(bool)val;
  if (Slow) period=400; else period=1;
  // 2. Files
  HKEY fkey;
  res=RegOpenKeyEx(skey,"Files",0,KEY_ALL_ACCESS,&fkey);
  if (res!=ERROR_SUCCESS) {RegCloseKey(skey); settingschanged=true; return; }
  LONG retcode=ERROR_SUCCESS; int i=0;
  while (retcode==ERROR_SUCCESS)
  { char ValName[MAX_PATH]; DWORD maxValName=MAX_PATH; DWORD type;
    char Val[MAX_PATH]; DWORD maxVal=MAX_PATH;
    retcode=RegEnumValue(fkey,i,(LPTSTR)ValName,&maxValName,NULL,&type,(LPBYTE)Val,&maxVal);
    if (retcode==ERROR_SUCCESS)
    { char fn[MAX_PATH]; strcpy(fn,Val);
      char pn[MAX_PATH]; strcpy(pn,&Val[strlen(Val)+1]);
      files->Insert(fn,pn);
    }
    i=i+1;
  }
  RegCloseKey(fkey);
  RegCloseKey(skey);
}


void TGhostsData::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Ghosts",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. Screen
  val=Screen;
  RegSetValueEx(skey,"Screen",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 3. Slow
  val=Slow;
  RegSetValueEx(skey,"Slow",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 2. files
  if (files==NULL) {RegCloseKey(skey); return;}
  RegDeleteKey(skey,"Files");
  HKEY fkey;
  res=RegCreateKeyEx(skey,"Files",0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&fkey,&disp);
  if (res!=ERROR_SUCCESS) {RegCloseKey(skey); return;}
  TFileData *f=files;
  int i=0;
  while (f!=NULL)
  { char c[15];
    lstrcpyn(c,f->path,15);
    c[10]='\0';
    if (strcmp(c,"<internal>")!=0)
    { char str[500];
      strcpy(str,f->name);
      int len=strlen(str);
      strcpy(&str[len+1],f->path);
      len=len+strlen(f->path)+2;
      char valname[30];
      wsprintf(valname,"file%i",i);
      i++;
      RegSetValueEx(fkey,valname,0,REG_BINARY,(CONST BYTE*)str,len);
    }
    f=f->next;
  }
  RegCloseKey(fkey);
  RegCloseKey(skey);
}



BOOL TGhostsData::GetControls(HWND hwnd)
{ if (ignorechanges) return FALSE;
  BOOL ischanged=FALSE;
  LRESULT res=SendDlgItemMessage(hwnd,ID_SCREEN,CB_GETCURSEL,0,0);
  if (res!=Screen) ischanged=TRUE; Screen=res;
  if (Screen<0) Screen=0;
  if (Screen>5) Screen=5;
  bool b=IsDlgButtonChecked(hwnd,ID_SLOW);
  if (b!=Slow) ischanged=TRUE; Slow=b;
  if (ischanged) settingschanged=true;
  return ischanged;
}

void TGhostsData::Add(HWND hwnd)
{ OPENFILENAME ofn;
  ofn.lStructSize=sizeof(ofn);
  ofn.hwndOwner=hwnd;
  ofn.hInstance=NULL; // ignored
  char filter[]="Bitmaps\0*.bmp\0\0";
  ofn.lpstrFilter=filter;
  ofn.lpstrCustomFilter=NULL;
  ofn.nMaxCustFilter=0; // ignored
  int index=1;
  ofn.nFilterIndex=index;
  char file[2000]; for (int i=0; i<2000; i++) file[i]='\0';
  ofn.lpstrFile=file;
  ofn.nMaxFile=500;
  char title[500];
  ofn.lpstrFileTitle=title;
  ofn.nMaxFileTitle=500;
  ofn.lpstrInitialDir=NULL;
  ofn.lpstrTitle=NULL;
  ofn.Flags=OFN_ALLOWMULTISELECT | OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
  ofn.nFileOffset=0; // ignored
  ofn.nFileExtension=0; // ignored
  char defext[20]; strcpy(defext,"bmp");
  ofn.lpstrDefExt=defext;
  ofn.lCustData=NULL; // ignored
  ofn.lpfnHook=NULL; // ignored
  ofn.lpTemplateName=NULL; // ignored
  BOOL res=GetOpenFileName(&ofn);
  if (!res) return;
  char tf[500];
  char root[500]; strcpy(root,"");
  int dest=0;
  int src=0;
  bool done=false;
  bool isinfile=false;
  bool doneroot=false;
  int numfiles=0;
  while (!done)
  { if (file[src]!='\0')
    { if (!isinfile) {isinfile=true; dest=0;}
      tf[dest]=file[src]; dest++; src++;
    }
    else
    { if (isinfile)
      { tf[dest]='\0';
        if (!doneroot)
        { strcpy(root,tf); doneroot=true; isinfile=false;
        }
        else
        { char fn[MAX_PATH]; strcpy(fn,tf);
          char pn[MAX_PATH]; strcpy(pn,root); strcat(pn,"\\"); strcat(pn,tf);
          files->Insert(fn,pn);
          numfiles++;
          isinfile=false;
        }
      }
      src++;
      if (file[src]=='\0') done=true;
    }
  }
  if (numfiles==0)
  { char fn[MAX_PATH]; strcpy(fn,title);
    char pn[MAX_PATH]; strcpy(pn,tf);
    files->Insert(fn,pn);
  }
  settingschanged=true;
  ShowFiles(hwnd);
  SendDlgItemMessage(hwnd,ID_LIST,LB_SETCURSEL,0,0);
  SelChange(hwnd);
}

void TGhostsData::Remove(HWND hwnd)
{ int res=SendDlgItemMessage(hwnd,ID_LIST,LB_GETCURSEL,0,0);
  int i=res;
  if (res==LB_ERR) return;
  settingschanged=true;
  if (res==0)
  { files=files->next;
    ShowFiles(hwnd);
    SendDlgItemMessage(hwnd,ID_LIST,LB_SETCURSEL,0,0);
    SelChange(hwnd);
    return;
  }
  TFileData *f=files;
  while (res>1)
  { if (f->next!=NULL) f=f->next;
    res=res-1;
  }
  f->next=f->next->next;
  ShowFiles(hwnd);
  SendDlgItemMessage(hwnd,ID_LIST,LB_SETCURSEL,i-1,0);
  SelChange(hwnd);
}

void TGhostsData::SelChange(HWND hwnd)
{ int res=SendDlgItemMessage(hwnd,ID_LIST,LB_GETCURSEL,0,0);
  if (res==LB_ERR) return;
  TFileData *f=files;
  while (res>0)
  { if (f->next!=NULL) f=f->next;
    res=res-1;
  }
  char c[15];
  lstrcpyn(c,f->path,15);
  c[10]='\0';
  bool isinternal=false;
  int internalid;
  if (strcmp(c,"<internal>")==0) isinternal=true;
  if (isinternal)
  { EnableWindow(GetDlgItem(hwnd,ID_REMOVE),false);
    int res=sscanf(&f->path[10],"%i",&internalid);
    if (res==0) isinternal=false;
  }
  else EnableWindow(GetDlgItem(hwnd,ID_REMOVE),true);
  previewbits.Reset();
  //
  HCURSOR hprevc=SetCursor(LoadCursor(NULL,IDC_WAIT));
  TBits loadbits;
  if (isinternal) loadbits.CreateDibFromResource(NULL,internalid);
  else loadbits.CreateDibFromFile(NULL,f->path);
  for (int i=0; i<256; i++)
  { BYTE b=(BYTE)(i&0xE0);
    previewbits.LogPal.Entries[i].peRed=b;
    previewbits.LogPal.Entries[i].peGreen=b;
    previewbits.LogPal.Entries[i].peBlue=b;
  }
  previewbits.ChangePalette();
  previewbits.EnableSubclass();
  previewbits.DibPainter.ProportionStretchToWindow();
  HWND hprevwnd=GetDlgItem(hwnd,ID_MONITOR);
  hprevwnd=GetWindow(hprevwnd,GW_CHILD);
  previewbits.CreateDibResample(hprevwnd,loadbits.surface[0],0,0,8,TRUE);
  TBitsSurface *bs=previewbits.LockSurface(0);
  if (bs!=NULL)
  { LPBYTE addr=bs->data.lpb;
    for (int i=0; i<previewbits.numbytes; i++)
    { BYTE b=*addr;
      b=(BYTE)(b&0xE0);
      if (b==0) b=0x20;
      if (b==0xE0) b=0xC0;
      *addr=b;
    }
  }
  previewbits.UnlockSurface(bs);
  previewbits.Flip();
  SetCursor(hprevc);
}

void TGhostsData::SetControls(HWND hwnd)
{ ignorechanges=TRUE;
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Normal screen");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 320x200");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 320x240");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 640x480");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 800x600");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_ADDSTRING,0,(LPARAM)"Fast 1024x768");
  SendDlgItemMessage(hwnd,ID_SCREEN,CB_SETCURSEL,Screen,0);
  UINT c=0; if (Slow) c=1;
  CheckDlgButton(hwnd,ID_SLOW,c);
  ShowFiles(hwnd,false);
  SendDlgItemMessage(hwnd,ID_LIST,LB_SETCURSEL,0,0);
  ignorechanges=FALSE;
  SelChange(hwnd);
}

void TGhostsData::ShowFiles(HWND hwnd,bool managechange)
{ if (managechange) ignorechanges=TRUE;
  int res=0;
  while (res!=LB_ERR) res=SendDlgItemMessage(hwnd,ID_LIST,LB_DELETESTRING,0,0);
  TFileData *f=files;
  while (f!=NULL)
  { SendDlgItemMessage(hwnd,ID_LIST,LB_ADDSTRING,0,(LPARAM)f->name);
    f=f->next;
  }
  if (managechange) ignorechanges=FALSE;
}




void TGhostsData::WinStop(HWND)
{ BitsDebug("Stop...");
  bits.Reset();
  SetEvent(abortworkevent);
  if (hthread!=0) WaitForSingleObject((HANDLE)hthread,10000);
  BitsDebug("Stopped.");
}

void TGhostsData::WinStart(HWND hwnd)
{ BitsDebug("Starting...");
  ResetEvent(abortworkevent);
  mode=0;
  currentimage=files;
  ischangingmode=true;
  switch (Screen)
  { case 1: { cx=320; cy=200; wantddx=true; } break;
    case 2: { cx=320; cy=240; wantddx=true; } break;
    case 3: { cx=640; cy=480; wantddx=true; } break;
    case 4: { cx=800; cy=600; wantddx=true; } break;
    case 5: { cx=1024;cy=768; wantddx=true; } break;
    default:
    { cx=GetSystemMetrics(SM_CXSCREEN); cy=GetSystemMetrics(SM_CYSCREEN);
      wantddx=false;
    }
  }
  if (!isfull) wantddx=false;
  if (isfull) bits.DibPainter.PositionAtPoint(0,0);
  else bits.DibPainter.StretchToWindow();
  bits.EnableSubclass(true,NULL);
  HRESULT err=bits.Create(hwnd,wantddx,2,cx,cy,8,true);
  if (wantddx) hasjustchangedmode=true;
  if (err!=BITS_OK) {BitsDebug("Error creating image."); ischangingmode=false; return;}
  OutputDebugString("CHANGED MODE!");
  ischangingmode=false;
  isstarted=true;
}


void PrepareThreadProc(void *lpvoid)
{ TGhostsData *g=(TGhostsData *)lpvoid;
  EnterCriticalSection(&g->facecritsec);
  HANDLE hThread=(HANDLE)g->hthread;
  SetThreadPriority(hThread,THREAD_PRIORITY_BELOW_NORMAL);
  //bool islowpriority=true;
  //DWORD starttime=timeGetTime();
  char c[15];
  lstrcpyn(c,g->currentimage->path,15);
  c[10]='\0';
  bool isinternal=false;
  int internalid;
  if (strcmp(c,"<internal>")==0) isinternal=true;
  if (isinternal)
  { int res=sscanf(&g->currentimage->path[10],"%i",&internalid);
    if (res==0) isinternal=false;
  }
  g->facebits.Reset();
  //
  TBits loadbits;
  if (isinternal) loadbits.CreateDibFromResource(NULL,internalid);
  else loadbits.CreateDibFromFile(NULL,g->currentimage->path);
  DWORD res=WaitForSingleObject(g->abortworkevent,0);
  if (res==WAIT_OBJECT_0) return;
  for (int i=0; i<256; i++)
  { BYTE b=(BYTE)(i&0xE0);
    g->facebits.LogPal.Entries[i].peRed=b;
    g->facebits.LogPal.Entries[i].peGreen=b;
    g->facebits.LogPal.Entries[i].peBlue=b;
  }
  g->facebits.ChangePalette();
  int acx=loadbits.cx, acy=loadbits.cy;
  acx=acx & 0xFFFC; // multiple of 4
  g->facebits.CreateDibResample(NULL,loadbits.surface[0],acx,acy,8,TRUE);
  TBitsSurface *bs=g->facebits.LockSurface(0);
  if (bs!=NULL)
  { LPBYTE addr=bs->data.lpb;
    for (int i=0; i<g->facebits.numbytes; i++)
    { BYTE b=*addr;
      b=(BYTE)(b&0xE0);
      if (b==0) b=0x20;
      if (b==0xE0) b=0xC0;
      *addr=b;
      addr++;
    }
  }
  g->facebits.UnlockSurface(bs);

  res=WaitForSingleObject(g->abortworkevent,0);
  if (res==WAIT_OBJECT_0) return;
  if (g->wantddx) g->facebits.surface[0].InvertBits();
  g->currentimage=g->currentimage->next; if (g->currentimage==NULL) g->currentimage=g->files;
  LeaveCriticalSection(&g->facecritsec);
  SetEvent(g->donefaceevent);
}


void TGhostsData::WinPrepare()
{ ResetEvent(donefaceevent);
  unsigned int time=timeGetTime();
  startfadeintime=time+25000;
  startfadeouttime=startfadeintime+10000;
  startnewimagetime=startfadeouttime+10000;
  hthread=_beginthread(PrepareThreadProc,4096,this);
  mode=1;
}


unsigned int TGhostsData::WinTimer(HWND)
{ if (issuspended) return period;
  BitsDebug("Timer...");
  if (mode==0) WinPrepare();
  if (mode==1)
  { DWORD res=WaitForSingleObject(donefaceevent,0);
    if (res!=WAIT_TIMEOUT)
    { ResetEvent(donefaceevent);
      WaitForSingleObject((HANDLE)hthread,10000);
      hthread=0;
      mode=2;
    }
  }
  WinDoFrame();
  unsigned int time=timeGetTime();
  if (mode==2)
  { if (time>=startfadeintime)
    { 
      unsigned int delay=time-startfadeintime;
      startfadeintime=startfadeintime+delay;
      startfadeouttime=startfadeouttime+delay;
      startnewimagetime=startnewimagetime+delay;
      screentop=0; screenbottom=bits.cy;
      screenleft=0; screenright=bits.cx;
      unsigned int rand=NEXTRAND3(time);
      int placewidth=screenright-facebits.cx; if (placewidth<1) placewidth=1;
      int placeheight=screenbottom-facebits.cy; if (placeheight<1) placeheight=1;
      int placex=rand % placewidth;
      rand=NEXTRAND3(rand);
      int placey=rand % placeheight;
      placex=placex & 0xFFFFFC;
      placey=placey & 0xFFFFFC;
      imagetop=placey; imageleft=placex;
      imageright=imageleft+facebits.cx; imagerestofline=0;
      if (imageright>screenright)
      { imagerestofline=imageright-screenright;
        imageright=screenright;
      }
      imagebottom=imagetop+facebits.cy; if (imagebottom>screenbottom) imagebottom=screenbottom;
      mode=3;
    }
  }
  if (mode==3)
  { if (time>=startfadeouttime) { mode=4;} }
  if (mode==4)
  { if (time>=startnewimagetime) { mode=0;} }
  return period;
}


void TGhostsData::WinDoFrame()
{ bool haschanged=false;
  DWORD rand; //=NEXTRAND3(timeGetTime());
  if (mode==1 || mode==2)
  { TBitsSurface *bs=bits.LockSurface(1);
    if (bs!=NULL)
    { haschanged=true;
      LPDWORD addr=bs->data.lpdw;
      int width=bits.cx/4;
      for (int y=0; y<bits.cy; y++)
      { rand=NEXTRAND(::rand()); rand=NEXTRAND2(rand);
        { for (int x=0; x<width; x++)
          { rand=NEXTRAND(rand);
            *addr++=(rand&0x1F1F1F1F) | 0xC0C0C0C0;
          }
        }
      }
    }
    bits.UnlockSurface(bs);
  }
  if (mode==3 || mode==4)
  { TBitsSurface *bs=bits.LockSurface(1);
    TBitsSurface *fs=facebits.LockSurface(0);
    if (bs!=NULL && fs!=NULL)
    { haschanged=true;
      LPDWORD baddr=bs->data.lpdw;
      LPDWORD faddr=fs->data.lpdw;
      for (int y=screentop; y<imagetop; y++)
      { rand=NEXTRAND(::rand()); rand=NEXTRAND2(rand);
        for (int x=screenleft; x<screenright; x=x+4)
        { rand=NEXTRAND(rand);
          *baddr=(rand&0x1F1F1F1F) | 0xC0C0C0C0;
          baddr++;
        }
      }
      for (int y=imagetop; y<imagebottom; y++)
      { rand=NEXTRAND(::rand()); rand=NEXTRAND2(rand);
        for (int x=screenleft; x<imageleft; x=x+4)
        { rand=NEXTRAND(rand);
          *baddr=(rand&0x1F1F1F1F) | 0xC0C0C0C0;
          baddr++;
        }
        for (int x=imageleft; x<imageright; x=x+4)
        { rand=NEXTRAND(rand);
          *baddr=(rand & 0x1F1F1F1F) | *faddr;
          baddr++; faddr++;
        }
        faddr=faddr+(imagerestofline/4);
        for (int x=imageright; x<screenright; x=x+4)
        { rand=NEXTRAND(rand);
          *baddr=(rand&0x1F1F1F1F) | 0xC0C0C0C0;
          baddr++;
        }
      }
      for (int y=imagebottom; y<screenbottom; y++)
      { rand=NEXTRAND(::rand()); rand=NEXTRAND2(rand);
        for (int x=screenleft; x<screenright; x=x+4)
        { rand=NEXTRAND(rand);
          *baddr=(rand&0x1F1F1F1F) | 0xC0C0C0C0;
          baddr++;
        }
      }
    }
    bits.UnlockSurface(bs);
    facebits.UnlockSurface(fs);
  }
  if (!haschanged) return;
  BYTE prob[9];
  int min,max;
  int time=timeGetTime();
  _fpreset();
  double fshow=0.0;
  switch (mode)
  { case 1: case 2: fshow=0.0; break;
    case 3: fshow=(((double)time)-((double)startfadeintime))/(((double)startfadeouttime)-((double)startfadeintime)); break;
    case 4: fshow=1.0-(((double)time)-((double)startfadeouttime))/(((double)startnewimagetime)-((double)startfadeouttime)); break;
  }
  if (fshow<0.0) fshow=0.0; if (fshow>1.0) fshow=1.0;
  fshow=fshow*fshow;
  min=128-fshow*50;
  max=128;
  for (int i=1; i<7; i++)
  { double f=((double)i)/6.0;
    prob[i]=(BYTE)(min+(int)(f*((double)max-min)));
    BYTE p=prob[i];
    int col=i*32;
    for (int off=0; off<32; off++,col++)
    { BYTE r=p;
      bits.LogPal.Entries[col].peRed=r;
      bits.LogPal.Entries[col].peGreen=(BYTE)(off*8);
      bits.LogPal.Entries[col].peBlue= (BYTE)(off*8);
    }
  }
  bits.Flip();
  bits.ChangePalette();
}

void TGhostsData::WinSuspend(HWND hwnd)
{ if (issuspended) return;
  if (!bits.isddx) return;
  issuspended=true;
  TBits suspendbits;
  suspendbits.CreateDibCopy(hwnd,bits.surface[1]);
  ischangingmode=true;
  bits.Reset();
  UINT oldval;
  SystemParametersInfo(SPI_SCREENSAVERRUNNING,1,&oldval,0);
  // because leaving DirectDraw would have suddenly enabled
  // ctrl-alt-delete and task switching. We want to disable it again.
  int scx=GetSystemMetrics(SM_CXSCREEN);
  int scy=GetSystemMetrics(SM_CYSCREEN);
  MoveWindow(hwnd,0,0,scx,scy,false);
  backdropbits.Reset();
  hasjustchangedmode=true;
  backdropbits.EnableFullSubclass(false);
  backdropbits.CreateDibCopy(hwnd,suspendbits.surface[0]);
  ischangingmode=false;
  suspendbits.Reset();
}

void TGhostsData::WinResume(HWND hwnd)
{ if (!issuspended) return;
  bits.Reset();
  if (isfull) bits.DibPainter.PositionAtPoint(0,0);
  else bits.DibPainter.StretchToWindow();
  bits.EnableSubclass(true,NULL);
  bits.CopyPaletteFrom(backdropbits);
  ischangingmode=true;
  backdropbits.Reset();
  HRESULT err=bits.Create(hwnd,wantddx,2,cx,cy,8,true);
  if (err!=BITS_OK) {ischangingmode=false; return;}
  if (wantddx) hasjustchangedmode=true;
  bits.Flip();
  ischangingmode=false;
  issuspended=false;
}


LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TGhostsData *sd=(TGhostsData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (sd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg)
  { case WM_CREATE:
    { BOOL isfull=false; if (GetParent(hwnd)==NULL) isfull=true;
      TGhostsData *sd=new TGhostsData(isfull);
      sd->ReadRegistry();
      SetWindowLong(hwnd,GWL_USERDATA,(LONG)sd);
    } break;
    case WM_ACTIVATE: case WM_NCACTIVATE:
    { if (sd->ischangingmode) return DefWindowProc(hwnd,msg,wParam,lParam);
    } break;
    case WM_MOUSEMOVE:
    { if (sd->ischangingmode || sd->hasjustchangedmode)
      { sd->hasjustchangedmode=false; return DefWindowProc(hwnd,msg,wParam,lParam);}
    } break;
    case SCRM_TIMER: {if (!sd->isstarted) sd->WinStart(hwnd); return sd->WinTimer(hwnd); }
    case SCRM_DIALOG:
    { if ((BOOL)wParam) sd->WinSuspend(hwnd); else sd->WinResume(hwnd); } break;
    case WM_DESTROY:
	 { sd->WinStop(hwnd);
      SetWindowLong(hwnd,GWL_USERDATA,0);
      delete sd;
    } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



BOOL WINAPI MyScreenSaverConfigureDialog(HWND hwnd, UINT msg,WPARAM wParam,LPARAM lParam)
{ TGhostsData *sd=(TGhostsData *)GetWindowLong(hwnd,DWL_USER);
  switch (msg)
  { case WM_INITDIALOG:
    { TGhostsData *sd=new TGhostsData(false);
      sd->ReadRegistry();
      SetWindowLong(hwnd,DWL_USER,(LONG)sd);
      sd->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { delete sd;
      SetWindowLong(hwnd,DWL_USER,0);
      return TRUE;
    }
    case WM_COMMAND:
    { int wID=LOWORD(wParam);
      int code=HIWORD(wParam);
      if (wID==ID_ADD && code==BN_CLICKED) sd->Add(hwnd);
      if (wID==ID_REMOVE && code==BN_CLICKED) sd->Remove(hwnd);
      if (wID==ID_LIST && code==LBN_SELCHANGE) sd->SelChange(hwnd);
      if (wID==ID_SCREEN && code==CBN_SELCHANGE) sd->GetControls(hwnd);
      if (wID==ID_SLOW && code==BN_CLICKED) sd->GetControls(hwnd);
    } break;
    case WM_NOTIFY:
    { LPNMHDR nmh=(LPNMHDR) lParam;
      UINT code=nmh->code;
      switch (code) {
        case (PSN_APPLY):
        { if (sd->settingschanged) sd->WriteRegistry();
          return TRUE;
        }
        case PSN_RESET:
        { return TRUE;
        }
        default: return FALSE;
      }
    }
  }
  return FALSE;
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  BitsInit(hInstance);
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=FALSE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
#ifdef STATICCODE
  strcpy(p->StaticCode,STATICCODE);
#endif
#ifdef DYNAMICCODE
  strcpy(p->DynamicCode,DYNAMICCODE);
#endif
  ScrSetPrefs(p);

  int res=ScrExecute(hInstance,MyScreenSaverProc,MyScreenSaverConfigureDialog);

  BitsClose();
  return res;
}



